// Component.h: interface for the Component class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_OMPONENT_H__3F88AECA_A2F5_44A0_A07D_73EF7E05D006__INCLUDED_)
#define AFX_OMPONENT_H__3F88AECA_A2F5_44A0_A07D_73EF7E05D006__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Identifier.h"

class BasicComponent;

class Component  
{
public:

	// where must the component start
	// to end at canonical origin [0, 0, TF]
	static void reverse(
		long targetX, long targetY, Direction targetDir,
		long * startX, long * startY, Direction * startDir)
	{
		switch(targetDir)
		{
		case TF:
			*startX = -targetX;
			*startY = -targetY;
			*startDir = TF;
			return;

		case TFX:
			*startX = -targetX;
			*startY = targetY;
			*startDir = TFX;
			return;

		case TR:
			*startX = -targetY;
			*startY = targetX;
			*startDir = TL;
			return;

		case TRX:
			*startX = -targetY;
			*startY = -targetX;
			*startDir = TRX;
			return;

		case TL:
			*startX = targetY;
			*startY = -targetX;
			*startDir = TR;
			return;

		case TLX:
			*startX = targetY;
			*startY = targetX;
			*startDir = TLX;
			return;

		case TB:
			*startX = targetX;
			*startY = targetY;
			*startDir = TB;
			return;

		case TBX:
			*startX = targetX;
			*startY = -targetY;
			*startDir = TBX;
			return;
		}
	}

	inline static void globalToLocal(
		long startX, long startY, Direction startDir, 
		long sxMin, long sxMax, long syMin, long syMax,
		long * xMin, long * xMax, long * yMin, long * yMax)
	{
		switch (startDir)
		{
		case TF:
			*xMin = sxMin - startX;
			*xMax = sxMax - startX;
			*yMin = syMin - startY;
			*yMax = syMax - startY;
			return;

		case TFX:
			*xMin = sxMin - startX;
			*xMax = sxMax - startX;
			*yMin = startY - syMax;
			*yMax = startY - syMin;
			return;

		case TR:
			*xMin = syMin - startY;
			*xMax = syMax - startY;
			*yMin = startX - sxMax;
			*yMax = startX - sxMin;
			return;

		case TRX:
			*xMin = syMin - startY;
			*xMax = syMax - startY;
			*yMin = sxMin - startX;
			*yMax = sxMax - startX;
			return;

		case TB:
			*xMin = startX - sxMax;
			*xMax = startX - sxMin;
			*yMin = startY - syMax;
			*yMax = startY - syMin;
			return;

		case TBX:
			*xMin = startX - sxMax;
			*xMax = startX - sxMin;
			*yMin = syMin - startY;
			*yMax = syMax - startY;
			return;

		case TL:
			*xMin = startY - syMax;
			*xMax = startY - syMin;
			*yMin = sxMin - startX;
			*yMax = sxMax - startX;
			return;

		case TLX:
			*xMin = startY - syMax;
			*xMax = startY - syMin;
			*yMin = startX - sxMax;
			*yMax = startX - sxMin;
			return;
		}
	}

	inline static void globalToLocal(
		long startX, long startY, Direction startDir, 
		long sx, long sy, 
		long * x, long * y)
	{
		switch (startDir)
		{
		case TF:
			*x = sx - startX;
			*y = sy - startY;
			return;

		case TFX:
			*x = sx - startX;
			*y = startY - sy;
			return;

		case TR:
			*x = sy - startY;
			*y = startX - sx;
			return;

		case TRX:
			*x = sy - startY;
			*y = sx - startX;
			return;

		case TB:
			*x = startX - sx;
			*y = startY - sy;
			return;

		case TBX:
			*x = startX - sx;
			*y = sy - startY;
			return;

		case TL:
			*x = startY - sy;
			*y = sx - startX;
			return;

		case TLX:
			*x = startY - sy;
			*y = startX - sx;
			return;
		}
	}

	inline static void localToGlobal(
		long startX, long startY, Direction startDir, 
		long sxMin, long sxMax, long syMin, long syMax,
		long * xMin, long * xMax, long * yMin, long * yMax)
	{
		switch (startDir)
		{
		case TF:
			*xMin = startX + sxMin;
			*xMax = startX + sxMax;
			*yMin = startY + syMin;
			*yMax = startY + syMax;
			return;

		case TFX:
			*xMin = startX + sxMin;
			*xMax = startX + sxMax;
			*yMin = startY - syMax;
			*yMax = startY - syMin;
			return;

		case TR:
			*xMin = startX - syMax;
			*xMax = startX - syMin;
			*yMin = startY + sxMin;
			*yMax = startY + sxMax;
			return;

		case TRX:
			*xMin = startX + syMin;
			*xMax = startX + syMax;
			*yMin = startY + sxMin;
			*yMax = startY + sxMax;
			return;

		case TB:			
			*xMin = startX - sxMax;
			*xMax = startX - sxMin;
			*yMin = startY - syMax;
			*yMax = startY - syMin;
			return;

		case TBX:
			*xMin = startX - sxMax;
			*xMax = startX - sxMin;
			*yMin = startY + syMin;
			*yMax = startY + syMax;
			return;

		case TL:
			*xMin = startX + syMin;
			*xMax = startX + syMax;
			*yMin = startY - sxMax;
			*yMax = startY - sxMin;
			return;

		case TLX:
			*xMin = startX - syMax;
			*xMax = startX - syMin;
			*yMin = startY - sxMax;
			*yMax = startY - sxMin;
			return;
		}
	}

	inline static void localToGlobal(
		long startX, long startY, Direction startDir, 
		long sx, long sy,
		long * x, long * y)
	{
		switch (startDir)
		{
		case TF:
			*x = startX + sx;
			*y = startY + sy;
			return;

		case TFX:
			*x = startX + sx;
			*y = startY - sy;
			return;

		case TR:
			*x = startX - sy;
			*y = startY + sx;
			return;

		case TRX:
			*x = startX + sy;
			*y = startY + sx;
			return;

		case TB:
			*x = startX - sx;
			*y = startY - sy;
			return;

		case TBX:
			*x = startX - sx;
			*y = startY + sy;
			return;

		case TL:
			*x = startX + sy;
			*y = startY - sx;
			return;

		case TLX:
			*x = startX - sy;
			*y = startY - sx;
			return;
		}
	}

	// what is the resulting direction if we put second after first
	inline static Direction combine(Direction first, Direction second) 
	{ 
		return s_combine[first * 8 + second]; 
	};

	// calculate second direction if combined with first makes result
	inline static Direction decombine(Direction first, Direction result) 
	{ 
		return s_decombine[first * 8 + result]; 
	};

	// calculate first direction if combined with second makes result
	inline static Direction subtract(Direction result, Direction second) 
	{ 
		return s_subtract[result * 8 + second]; 
	};

	inline static const char * strDir(Direction dir)
	{
		switch(dir)
		{
		case TF:
			return "F";
		case TFX:
			return "Fx";
		case TB:
			return "B";
		case TBX:
			return "Bx";
		case TR:
			return "R";
		case TRX:
			return "Rx";
		case TL:
			return "L";
		case TLX:
			return "Lx";
		default:
			return "?";
		}
	}

private:
	static Direction s_combine[];
	static Direction s_decombine[];
	static Direction s_subtract[];
};

#endif // !defined(AFX_OMPONENT_H__3F88AECA_A2F5_44A0_A07D_73EF7E05D006__INCLUDED_)
