// PointSet.cpp: implementation of the PointSet class.
//
//////////////////////////////////////////////////////////////////////

#include "PointSet.h"
#include <Windows.h>
#include <assert.h>
#include <stdio.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

PointSet::Block8k * PointSet::s_free8k = NULL;
PointSet::Block4k * PointSet::s_free4k = NULL;
PointSet::Block2k * PointSet::s_free2k = NULL;
PointSet::Block1k * PointSet::s_free1k = NULL;
PointSet::Block05k * PointSet::s_free05k = NULL;

long PointSet::stat_used8k = 0;
long PointSet::stat_used4k = 0;
long PointSet::stat_used2k = 0;
long PointSet::stat_used1k = 0;
long PointSet::stat_used05k = 0;

long PointSet::stat_free8k = 0;
long PointSet::stat_free4k = 0;
long PointSet::stat_free2k = 0;
long PointSet::stat_free1k = 0;
long PointSet::stat_free05k = 0;


PointSet::PointSet()
{
	for (long i = 0; i < 200; i++)
	{
		m_index[i].count = 0;
		m_index[i].maxCount = 0;
		m_index[i].pointer.p05k = NULL;
	}
}

PointSet::~PointSet()
{
// I will not deallocate the structures - they're used once per program run
}

PointSet::Block8k * PointSet::getBlock8k()
{
	if (s_free8k == NULL)
	{
		MemoryBlock * mb = new MemoryBlock;
		assert(mb != NULL);

		long i = sizeof(*mb) / sizeof(mb->blk[0]);

		do
		{
			--i;
			mb->blk[i].next = s_free8k;
			s_free8k = &mb->blk[i];
			++stat_free8k;
		} while (i != 0);
	}
	Block8k * b = s_free8k;
	s_free8k = s_free8k->next;
	--stat_free8k;
	++stat_used8k;
	return b;
}

PointSet::Block4k * PointSet::getBlock4k()
{
	if (s_free4k != NULL)
	{
		Block4k * b = s_free4k;
		s_free4k = s_free4k->next;
		--stat_free4k;
		++stat_used4k;
		return b;
	}
	else
	{
		Block8k * b = getBlock8k();
		Block4k * bb = &b->blk[1];
		bb->next = NULL;
		s_free4k = bb;
		--stat_used8k;
		++stat_free4k;
		++stat_used4k;
		return &b->blk[0];
	}
}

PointSet::Block2k * PointSet::getBlock2k()
{
	if (s_free2k != NULL)
	{
		Block2k * b = s_free2k;
		s_free2k = s_free2k->next;
		--stat_free2k;
		++stat_used2k;
		return b;
	}
	else
	{
		Block4k * b = getBlock4k();
		Block2k * bb = &b->blk[1];
		bb->next = NULL;
		s_free2k = bb;
		--stat_used4k;
		++stat_free2k;
		++stat_used2k;
		return &b->blk[0];
	}
}

PointSet::Block1k * PointSet::getBlock1k()
{
	if (s_free1k != NULL)
	{
		Block1k * b = s_free1k;
		s_free1k = s_free1k->next;
		--stat_free1k;
		++stat_used1k;
		return b;
	}
	else
	{
		Block2k * b = getBlock2k();
		Block1k * bb = &b->blk[1];
		bb->next = NULL;
		s_free1k = bb;
		--stat_used2k;
		++stat_free1k;
		++stat_used1k;
		return &b->blk[0];
	}
}

PointSet::Block05k * PointSet::getBlock05k()
{
	if (s_free05k != NULL)
	{
		Block05k * b = s_free05k;
		s_free05k = s_free05k->next;
		--stat_free05k;
		++stat_used05k;
		return b;
	}
	else
	{
		Block1k * b = getBlock1k();
		Block05k * bb = &b->blk[1];
		bb->next = NULL;
		s_free05k = bb;
		--stat_used1k;
		++stat_free05k;
		++stat_used05k;
		return &b->blk[0];
	}
}

void PointSet::putBlock8k(Block8k * b)
{
	b->next = s_free8k;
	s_free8k = b;
	--stat_used8k;
	++stat_free8k;
}

void PointSet::putBlock4k(Block4k * b)
{
	b->next = s_free4k;
	s_free4k = b;
	--stat_used4k;
	++stat_free4k;
}

void PointSet::putBlock2k(Block2k * b)
{
	b->next = s_free2k;
	s_free2k = b;
	--stat_used2k;
	++stat_free2k;
}

void PointSet::putBlock1k(Block1k * b)
{
	b->next = s_free1k;
	s_free1k = b;
	--stat_used1k;
	++stat_free1k;
}

void PointSet::putBlock05k(Block05k * b)
{
	b->next = s_free05k;
	s_free05k = b;
	--stat_used05k;
	++stat_free05k;
}


void PointSet::addPoint(long x, long y, Direction dir)
{
	x = (x + 640) % 1280;
	long x1 = x / 256;

	y = (y + 640) % 1280;
	long y1 = y / 256;

	unsigned short v = ((y & 255) << 8) | (x & 255);

	MapHeader * mh = &m_index[dir * 25 + y1 * 5 + x1];

	if (mh->pointer.p05k != NULL)
	{
		// block found
		if (mh->maxCount != 65536)
		{
			// value list
			// try to find the value first
			long l = 0;
			long h = mh->count;

			while (h > l)
			{
				long c = (l + h) / 2;
				unsigned short pt = mh->pointer.p4k->data[c];
				if (pt <= v)
				{
					if (pt != v)
					{
						l = c + 1;
					}
					else
					{
						// value found
						return;
					}
				}
				else
				{
					h = c;
				}
			}

			// make sure once more
			if ((l < mh->count) && (mh->pointer.p4k->data[l] == v))
			{
				return;
			}

			// value not found
			// check block size

			if (mh->count == mh->maxCount)
			{
				switch (mh->count)
				{
				case c05k:
					// enlarging block from 0.5k to 1k
					{
						Block1k * b = getBlock1k();
						memcpy(b->data, mh->pointer.p05k->data, sizeof(Block05k));
						putBlock05k(mh->pointer.p05k);
						mh->pointer.p1k = b;
						mh->maxCount = c1k;
						break;
					}
				case c1k:
					// enlarging block from 1k to 2k
					{
						Block2k * b = getBlock2k();
						memcpy(b->data, mh->pointer.p1k->data, sizeof(Block1k));
						putBlock1k(mh->pointer.p1k);
						mh->pointer.p2k = b;
						mh->maxCount = c2k;
						break;
					}
				case c2k:
					// enlarging block from 2k to 4k
					{
						Block4k * b = getBlock4k();
						memcpy(b->data, mh->pointer.p2k->data, sizeof(Block2k));
						putBlock2k(mh->pointer.p2k);
						mh->pointer.p4k = b;
						mh->maxCount = c4k;
						break;
					}
				case c4k:
					// converting the block to bitmap
					{
						Block8k * b = getBlock8k();
						// clear the bitmap
						long i = c8k;
						unsigned long * lp = &b->map[0];
						do
						{
							*lp = 0;
							++lp;
							--i;
						} while (i != 0);

						// put all points to it
						unsigned short * p = &mh->pointer.p4k->data[0];
						i = mh->count;
						do
						{
							unsigned short w = *p;
							++p;

							b->map[w >> 5] |= 1 << (w & 31);

							--i;
						} while (i != 0);

						// and put the 'v' value too

						b->map[v >> 5] |= 1 << (v & 31);

						// swap pages

						putBlock4k(mh->pointer.p4k);
						mh->pointer.p8k = b;
						mh->maxCount = 65536;

						// now return, everything is done

						return;
					}
				default:
					// this shouldn't happen
					assert(false);
				}
			}

			// position found and ensured that there is enough place

			unsigned short * p = &mh->pointer.p4k->data[l];
			l = mh->count - l;
			if (l != 0)
			{
				do
				{
					unsigned short w = *p;
					*p = v;
					v = w;
					++p;
					--l;
				} while (l != 0);
			}
			*p = v;
			++mh->count;

			// done
		}
		else
		{
			// bitmap
			unsigned long b = 1 << (v & 31);
			unsigned long * p = &mh->pointer.p8k->map[v >> 5];
			if ((*p & b) == 0)
			{
				// just place the bit
				++mh->count;
				*p |= b;
			}
		}
	}
	else
	{
		// pointer is null; allocate new block
		mh->pointer.p05k = getBlock05k();
		mh->pointer.p05k->data[0] = v;
		mh->count = 1;
		mh->maxCount = c05k;
		return;
	}
}

bool PointSet::checkPoint(long x, long y, Direction dir)
{
	x = (x + 640) % 1280;
	long x1 = x / 256;

	y = (y + 640) % 1280;
	long y1 = y / 256;

	unsigned short v = ((y & 255) << 8) | (x & 255);

	MapHeader * mh = &m_index[dir * 25 + y1 * 5 + x1];

	if (mh->pointer.p05k == NULL)
	{
		return false;
	}

	if (mh->maxCount != 65536)
	{
		long l = 0;
		long h = mh->count;

		while (h > l)
		{
			long c = (l + h) / 2;
			unsigned short pt = mh->pointer.p4k->data[c];
			if (pt <= v)
			{
				if (pt != v)
				{
					l = c + 1;
				}
				else
				{
					return true;
				}
			}
			else
			{
				h = c;
			}
		}

		return ((l < mh->count) && (mh->pointer.p4k->data[l] == v));
	}
	else
	{
		// bitmap
		return ((((mh->pointer.p8k->map[v >> 5]) >> (v & 31)) & 1) != 0);
	}
}
/*
	long m_passNdx;
	long m_passX;
	long m_passY;
	Direction m_passDir;
	long m_passDetailNdx;
	unsigned long m_passDetailBit;


	x = (x + 640) % 1280;
	long x1 = x / 256;

	y = (y + 640) % 1280;
	long y1 = y / 256;

	unsigned short v = ((y & 255) << 8) | (x & 255);

	MapHeader * mh = &m_index[dir * 25 + y1 * 5 + x1];
*/
void PointSet::makePassXYD()
{
	long n = m_passNdx;
	m_passX = (n % 5) * 256 - 640;
	n /= 5;
	m_passY = (n % 5) * 256 - 640;
	n /= 5;
	m_passDir = (Direction) (n % 8);
}

bool PointSet::getFirstDetail(MapHeader * mh, unsigned short & v)
{
	if (mh->count == 0)
	{
		return false;
	}

	if (mh->maxCount != 65536)
	{
		m_passDetailNdx = 0;
		v = mh->pointer.p4k->data[0];
		return true;
	}
	else
	{
		m_passDetailNdx = 0;
		unsigned long * p = &mh->pointer.p8k->map[0];
		while (*p == 0)
		{
			// there must be one...
			++p;
			++m_passDetailNdx;
		}
		m_passDetailBit = 1;
		m_passDetailBitNo = 0;
		unsigned long b = *p;
		while ((b & m_passDetailBit) == 0)
		{
			// again there must be at least one bit on
			m_passDetailBit <<= 1;
			++m_passDetailBitNo;
		}
		v = (unsigned short) ((m_passDetailNdx << 5) | m_passDetailBitNo);
		return true;
	}
}

bool PointSet::getNextDetail(MapHeader * mh, unsigned short & v)
{
	if (mh->maxCount != 65536)
	{
		// list
		long n = m_passDetailNdx + 1;
		if (n == mh->count)
		{
			return false;
		}
		else
		{
			m_passDetailNdx = n;
			v = mh->pointer.p4k->data[n];
			return true;
		}
	}
	else
	{
		// bitmap

		unsigned long * p = &mh->pointer.p8k->map[m_passDetailNdx];
		unsigned long w = *p;

		while (m_passDetailBitNo != 31)
		{
			++m_passDetailBitNo;
			m_passDetailBit <<= 1;
			if ((w & m_passDetailBit) != 0)
			{
				// found
				v = (unsigned short) ((m_passDetailNdx << 5) | m_passDetailBitNo);
				return true;
			}
		}

		// no more bits in current word
		// find another word

		do
		{
			++m_passDetailNdx;
			if (m_passDetailNdx == c8k)
			{
				// no more words
				return false;
			}
			++p;
			w = *p;
		} while (w == 0);

		// word found
		m_passDetailBit = 1;
		m_passDetailBitNo = 0;
		while ((w & m_passDetailBit) == 0)
		{
			m_passDetailBit <<= 1;
			++m_passDetailBitNo;
		}
		v = (unsigned short) ((m_passDetailNdx << 5) | m_passDetailBitNo);
		return true;
	}
}

bool PointSet::deleteAndNextDetail(MapHeader * mh, unsigned short & v)
{
	--mh->count;
	if (mh->maxCount != 65536)
	{
		// list
		unsigned short * p = &mh->pointer.p4k->data[m_passDetailNdx];
		long i = mh->count - m_passDetailNdx;
		if (i != 0)
		{
			// shift and take next one
			do
			{
				*p = *(p + 1);
				++p;
				--i;
			} while (i != 0);
			v = mh->pointer.p4k->data[m_passDetailNdx];
			return true;
		}
		else
		{
			// nothing to shift
			// equals no more values
			return false;
		}
	}
	else
	{
		// bitmap
		// delete the bit and continue as usual
		unsigned long * p = &mh->pointer.p8k->map[m_passDetailNdx];
		*p &= ~m_passDetailBit;
		return getNextDetail(mh, v);
	}
}


bool PointSet::getFirst(long &x, long &y, Direction &dir)
{
	m_passNdx = 0;
	do
	{
		unsigned short v;
		if (getFirstDetail(&m_index[m_passNdx], v))
		{
			// something found
			makePassXYD();
			x = (v & 255) + m_passX;
			y = (v >> 8) + m_passY;
			dir = m_passDir;
			return true;
		}
		++m_passNdx;
	} while (m_passNdx != 200);
	return false;
}

bool PointSet::getNext(long &x, long &y, Direction &dir)
{
	unsigned short v;
	if (!getNextDetail(&m_index[m_passNdx], v))
	{
		do
		{
			++m_passNdx;
			if (m_passNdx == 200)
			{
				return false;
			}
		} while (!getFirstDetail(&m_index[m_passNdx], v));
		makePassXYD();
	}
	x = (v & 255) + m_passX;
	y = (v >> 8) + m_passY;
	dir = m_passDir;
	return true;
}

bool PointSet::deleteAndNext(long &x, long &y, Direction &dir)
{
	unsigned short v;
	if (!deleteAndNextDetail(&m_index[m_passNdx], v))
	{
		do
		{
			++m_passNdx;
			if (m_passNdx == 200)
			{
				return false;
			}
		} while (!getFirstDetail(&m_index[m_passNdx], v));
		makePassXYD();
	}
	x = (v & 255) + m_passX;
	y = (v >> 8) + m_passY;
	dir = m_passDir;
	return true;
}

long PointSet::getCount()
{
	long s = 0;
	for (long i = 0; i < 200; ++i)
	{
		s += m_index[i].count;
	}
	return s;
}

void PointSet::printStatistics()
{
	printf("8.0k: %d blocks used, %d blocks free\n", stat_used8k, stat_free8k);
	printf("4.0k: %d blocks used, %d blocks free\n", stat_used4k, stat_free4k);
	printf("2.0k: %d blocks used, %d blocks free\n", stat_used2k, stat_free2k);
	printf("1.0k: %d blocks used, %d blocks free\n", stat_used1k, stat_free1k);
	printf("0.5k: %d blocks used, %d blocks free\n", stat_used05k, stat_free05k);
}