// Search.cpp: implementation of the Search class.
//
//////////////////////////////////////////////////////////////////////

#include <math.h>

#include "Search.h"
#include "BasicComponent.h"
#include "CompoundComponent.h"
#include "ComponentVariant.h"
#include "ComponentRegister.h"
#include "AcTree.h"
#include "AcTreeNode.h"
#include "Writer.h"
#include "PointSet.h"
#include "Progress.h"
#include <time.h>
#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Search::Search()
{
	m_register = NULL;
	m_componentList = NULL;
	m_componentListLength = 0;
	m_sorted = false;
	m_closed = true;
}

Search::~Search()
{
//	delete m_register;
}

void Search::test()
{
	m_register = new ComponentRegister("data/data.ini");
	assert(m_register != NULL);

	m_register->prepare(0, 0, false, false);

	Writer::print("Generating test pattern...\n");
	// there will be variant matrix,
	// and one example of each variant 
	long n = m_register->getNumVariants() * m_register->getNumVariants() * 2 + m_register->getNumVariants() * 2;
	m_cc = new CompoundComponent(n, m_register->getMaxVariants(), 0, "???");
	assert(m_cc != NULL);
	CompoundComponent * cc = new CompoundComponent(2, m_register->getMaxVariants(), 0, "???");
	assert(cc != NULL);

	// first find how big is the biggest component tuple
	long i;
	long j;
	BasicComponent * comp1;
	BasicComponent * comp2;
	ComponentVariant * var1;
	ComponentVariant * var2;

	long dx = 0;
	long dy = 0;

	for (i = 0; i < m_register->getNumVariants(); i++)
	{
		var1 = m_register->getVariantById(i);
		comp1 = m_register->getComponentById(var1->getComponentId());

		m_cc->addVariant(0, 0, TF, comp1, var1);
		for (j = 0; j < m_register->getNumVariants(); j++)
		{
			var2 = m_register->getVariantById(j);
			if (m_register->canFollow(var1, var2))
			{
				comp2 = m_register->getComponentById(var2->getComponentId());

				m_cc->addVariant(comp2, var2);

				long loX, hiX, loY, hiY;

				m_cc->getSize(loX, hiX, loY, hiY);

				hiX -= loX;
				hiY -= loY;
				if (dx < hiX)
				{
					dx = hiX;
				}
				if (dy < hiY)
				{
					dy = hiY;
				}

				m_cc->removeLastComponent();
			}
		}
		m_cc->removeLastComponent();
	}

	dx += 10;
	dy += 10;


	// and then put the whole matrix in
	for (i = -1; i < m_register->getNumVariants(); i++)
	{
		if (i >= 0)
		{
			var1 = m_register->getVariantById(i);
			comp1 = m_register->getComponentById(var1->getComponentId());
		}
		else
		{
			var1 = NULL;
			comp1 = NULL;
		}

		for (j = -1; j < m_register->getNumVariants(); j++)
		{

			if (j >= 0)
			{
				var2 = m_register->getVariantById(j);
				comp2 = m_register->getComponentById(var2->getComponentId());
			}
			else
			{
				var2 = NULL;
				comp2 = NULL;
			}

			if ((var1 == NULL) || (var2 == NULL) || m_register->canFollow(var1, var2))
			{
				while (cc->getNumComponents() > 0)
				{
					cc->removeLastComponent();
				}

				if (var1 != NULL)
				{
					cc->addVariant(0, 0, TF, comp1, var1);
					if (var2 != NULL)
					{
						cc->addVariant(comp2, var2);
					}
				}
				else if (var2 != NULL)
				{
					cc->addVariant(0, 0, TF, comp2, var2);
				}

				long loX, hiX, loY, hiY;
				cc->getSize(loX, hiX, loY, hiY);
				loX = (loX - hiX) / 2 - loX + dx * j;
				loY = (loY - hiY) / 2 - loY + dy * i;

				if (var1 != NULL)
				{
					m_cc->addVariant(loX, loY, TF, comp1, var1);
					if (var2 != NULL)
					{
						m_cc->addVariant(comp2, var2);
					}
				}
				else if (var2 != NULL)
				{
					m_cc->addVariant(loX, loY, TF, comp2, var2);
				}
			}
		}
	}

	m_cc->printRle(true);

	delete m_cc;
	delete cc;

	// and now display all invalid sequences

	while (NULL != (cc = m_register->getInvalidSequence()))
	{
		if (0 != cc->check(false))
		{
			cc->print(0);
			Writer::print(" - unnecessary sequence\n");
		}
		cc->printRle(true);
		delete cc;
	}

	Writer::close();
}

void Search::run(
	long period, 
	long compression,
	long maxCells,
	long maxResults,
	bool sorted,
	long treeSizeKB,
	long ignoreSets,
	bool useMemory)
{
	m_maxCells = maxCells;
	m_maxResults = maxResults;
	m_sorted = sorted;
	m_run = true;
	m_interrupted = false;
	m_ignoreSets = ignoreSets;
	m_optimizeMaxCells = m_sorted && (m_maxResults != 0);

	m_register = new ComponentRegister("data/data.ini");
	assert(m_register != NULL);

	// deactivate all components we're not going to use
	m_register->prepare(period, compression, m_closed, useMemory);

	// and build Aho/Corasick tree
	m_register->buildTree(treeSizeKB * 1024);

	// first collect number of components available
	long i;
	long n = 0;
	for (i = 0; i < m_register->getNumComponents(); i++)
	{
		if (m_register->getComponentById(i)->getActive())
		{
			n++;
		}
	}

	// move all available components to a list
	delete [] m_componentList;
	m_componentList = new BasicComponent * [n];
	assert(m_componentList != NULL);

	m_componentListLength = 0;

	for (i = 0; i < m_register->getNumComponents(); i++)
	{
		if (m_register->getComponentById(i)->getActive())
		{
			m_componentList[m_componentListLength] =
				m_register->getComponentById(i);
			m_componentListLength++;
		}
	}

	m_numSets = 0;
	if (m_componentListLength > 0)
	{
		if (useMemory)
		{
			runExtensiveWay();
		}
		else
		{
			runCheapWay();
		}
	}

	printSorted();

	if (m_interrupted)
	{
		if (m_numSets > 0)
		{
			char buf[80];
			sprintf(buf, "Search interrupted at set %d\n", m_numSets);
			Writer::print(buf);
		}
		else
		{
			Writer::print("Search interrupted\n");
		}
	}
	else
	{
		Writer::print("Search complete\n");
	}

	// now delete the array
	delete [] m_componentList;
	m_componentList = NULL;
	m_componentListLength = 0;

	Writer::close();
}

// memory-extensive search
void Search::runExtensiveWay()
{
	// at least one target is necessary
	if (m_targetList.empty())
	{
		return;
	}

	long maxDt;
	{
		TargetList::iterator it = m_targetList.end();
		--it;
		maxDt = (*it)->dt + 1;
	}

	long halfDt = maxDt / 2;
	// 0 .. halfDt - 1 will be filled 'forward'
	// halfDt .. maxDt - 1 will be filled 'backward'

	Writer::print("Creating search map...\n");

	PointSet * pointSet = new PointSet[maxDt];
	assert(pointSet != NULL);

	// insert the first point
	pointSet[0].addPoint(0, 0, TF);

	long beg_xlo, beg_xhi, beg_ylo, beg_yhi;

	if (m_maxCells != 0)
	{
		// calculate the 'beginning' size
		// first determine if all targets are closed
		bool allclosed = true;
		{
			// insert all targets over first half
			TargetList::iterator it = m_targetList.begin();
			while (allclosed && (it != m_targetList.end()))
			{
				// reverse the target - we will _start_ the search at it
				allclosed = (*it)->closed;
				++it;
			}
		}

		long bx1, bx2, by1, by2;
		long ex1, ex2, ey1, ey2;
		long x, y;
		Direction d;

		m_componentList[0]->getMinimumSize(bx1, bx2, by1, by2);
		Component::reverse(
			m_componentList[0]->getDx(), 
			m_componentList[0]->getDy(),
			m_componentList[0]->getDirection(),
			&x, &y, &d);
		m_componentList[0]->getMinimumSize(x, y, d, ex1, ex2, ey1, ey2);

		long i = m_componentListLength - 1;
		while (i != 0)
		{
			long xx1, xx2, yy1, yy2;
			m_componentList[i]->getMinimumSize(xx1, xx2, yy1, yy2);
			bx1 = (bx1 > xx1) ? bx1 : xx1;
			bx2 = (bx2 < xx2) ? bx2 : xx2;
			by1 = (by1 > xx1) ? by1 : xx1;
			by2 = (by2 < xx2) ? by2 : xx2;

			Component::reverse(
				m_componentList[i]->getDx(), 
				m_componentList[i]->getDy(),
				m_componentList[i]->getDirection(),
				&x, &y, &d);
			m_componentList[i]->getMinimumSize(x, y, d, xx1, xx2, yy1, yy2);
			ex1 = (ex1 > xx1) ? ex1 : xx1;
			ex2 = (ex2 < xx2) ? ex2 : xx2;
			ey1 = (ey1 > xx1) ? ey1 : xx1;
			ey2 = (ey2 < xx2) ? ey2 : xx2;

			--i;
		}

		if (allclosed)
		{
			ex1 = (ex1 < bx1) ? ex1 : bx1;
			ex2 = (ex2 > bx2) ? ex2 : bx2;
			ey1 = (ey1 < bx1) ? ey1 : bx1;
			ey2 = (ey2 > bx2) ? ey2 : bx2;
		}

		beg_xlo = ex1;
		beg_xhi = ex2;
		beg_ylo = ey1;
		beg_yhi = ey2;
	}

	Progress::activate(maxDt * 2);

	{
		for (long i = 0; i < halfDt; ++i)
		{
			Progress::step();

			PointSet * ps = &pointSet[i];

			long x;
			long y;
			Direction d;

			for (long c = 0; c < m_componentListLength; ++c)
			{
				BasicComponent * cc = m_componentList[c];
				long t = i + cc->getDt();
				if (t < halfDt)
				{
					if (ps->getFirst(x, y, d))
					{
						do
						{
							if (m_maxCells != 0)
							{
								long x1, x2, y1, y2;
								// calculate result size
								cc->getMinimumSize(
									x + cc->getRdx(d), 
									y + cc->getRdy(d), 
									cc->getRdir(d), 
									x1, x2, y1, y2);

								x1 = (x1 < beg_xlo) ? x1 : beg_xlo;
								x2 = (x2 > beg_xhi) ? x2 : beg_xhi;
								y1 = (y1 < beg_ylo) ? y1 : beg_ylo;
								y2 = (y2 > beg_yhi) ? y2 : beg_yhi;

								if (((x2 - x1) * (y2 - y1)) <= m_maxCells)
								{
									// size is OK
									pointSet[t].addPoint(
										x + cc->getRdx(d), 
										y + cc->getRdy(d), 
										cc->getRdir(d));
								}
							}
							else
							{
								// just insert it
								pointSet[t].addPoint(
									x + cc->getRdx(d), 
									y + cc->getRdy(d), 
									cc->getRdir(d));
							}
						} while (ps->getNext(x, y, d));
					}
				}
			}
		}
	}

	{
		// insert all targets over first half
		TargetList::iterator it = m_targetList.begin();
		while (it != m_targetList.end())
		{
			// reverse the target - we will _start_ the search at it
			Target * target = *it;
			long x;
			long y;
			Direction dir;

			if (target->dt >= halfDt)
			{
				Component::reverse(target->dx, target->dy, target->dir, &x, &y, &dir);
				pointSet[target->dt].addPoint(x, y, dir);
			}

			++it;
		}
	}

	// run the upper half from end to beginning
	{
		for (long i = maxDt - 1; i >= halfDt; --i)
		{
			Progress::step();

			PointSet * ps = &pointSet[i];

			long x;
			long y;
			Direction d;
			for (long c = 0; c < m_componentListLength; ++c)
			{
				BasicComponent * cc = m_componentList[c];
				long t = i - cc->getDt();
				if (t >= halfDt)
				{
					if (ps->getFirst(x, y, d))
					{
						do
						{
							if (m_maxCells != 0)
							{
								long x1, x2, y1, y2;
								// calculate result size
								cc->getMinimumSize(
									x, y, d, 
									x1, x2, y1, y2);

								x1 = (x1 < beg_xlo) ? x1 : beg_xlo;
								x2 = (x2 > beg_xhi) ? x2 : beg_xhi;
								y1 = (y1 < beg_ylo) ? y1 : beg_ylo;
								y2 = (y2 > beg_yhi) ? y2 : beg_yhi;

								if (((x2 - x1) * (y2 - y1)) <= m_maxCells)
								{
									// size is OK
									pointSet[t].addPoint(
										x + cc->getDx(d), 
										y + cc->getDy(d), 
										cc->getDirection(d));
								}
							}
							else
							{
								// just insert it
								pointSet[t].addPoint(
									x + cc->getDx(d), 
									y + cc->getDy(d), 
									cc->getDirection(d));
							}
						} while (ps->getNext(x, y, d));
					}
				}
			}
		}
	}

	// and now for the backwards run
	{
		for (long i = halfDt; i < maxDt; ++i)
		{
			Progress::step();

			PointSet * ps = &pointSet[i];

			long x;
			long y;
			Direction d;

			if (ps->getFirst(x, y, d))
			{
				bool cont;

				do
				{
					cont = false;

					long c = m_componentListLength;
					do
					{
						--c;
						BasicComponent * cc = m_componentList[c];
						long t = i - cc->getDt();
						if (t >= 0)
						{
							cont = pointSet[t].checkPoint(
								x + cc->getDx(d), 
								y + cc->getDy(d), 
								cc->getDirection(d));
						}
					} while (!cont && (c != 0));

					if (cont)
					{
						cont = ps->getNext(x, y, d);
					}
					else
					{
						cont = ps->deleteAndNext(x, y, d);
					}

				} while (cont);
			}
		}
	}

	// and the last round
	{
		for (long i = halfDt - 1; i >= 0; --i)
		{
			Progress::step();

			PointSet * ps = &pointSet[i];

			long x;
			long y;
			Direction d;

			if (ps->getFirst(x, y, d))
			{
				bool cont;

				do
				{
					cont = false;

					long c = m_componentListLength;
					do
					{
						--c;
						BasicComponent * cc = m_componentList[c];
						long t = i + cc->getDt();
						
						if (t < maxDt)
						{
							cont = pointSet[t].checkPoint(
								x + cc->getRdx(d), 
								y + cc->getRdy(d), 
								cc->getRdir(d));
						}
					} while (!cont && (c != 0));

					if (cont)
					{
						cont = ps->getNext(x, y, d);
					}
					else
					{
						cont = ps->deleteAndNext(x, y, d);
					}

				} while (cont);
			}

			// After reducing each level we must add all target points
			// for that level
			{
				TargetList::iterator it = m_targetList.begin();
				while (it != m_targetList.end())
				{
					Target * target = *it;
					if (target->dt == i)
					{
						long x;
						long y;
						Direction dir;
						// reverse the target - we will _start_ the search at it
						Component::reverse(target->dx, target->dy, target->dir, &x, &y, &dir);
						pointSet[i].addPoint(x, y, dir);
					}
					++it;
				}
			}
		}
	}

	Progress::cleanup();
	Writer::print("Search map created\n");
	// Everything prepared
	// let's do the backtracking

	m_numSets = 0;

	Writer::print("Counting component sets...\n");

	{
		TargetList::iterator it = m_targetList.begin();
		while (it != m_targetList.end())
		{
			searchTarget(*it, pointSet, false);
			++it;
		}
	}

	char buf[80];
	sprintf(buf, "Found %d component sets\n", m_numSets);
	Writer::print(buf);

	if (m_ignoreSets == 0)
	{
		Writer::print("Searching...\n");
	}
	else
	{
		sprintf(buf, "Searching (starting at set %d) ...\n", m_ignoreSets + 1);
		Writer::print(buf);
	}

	Progress::activate(m_numSets);

	long numSets = m_numSets;

	m_numSets = 0;

	{
		TargetList::iterator it = m_targetList.begin();
		while (m_run && (it != m_targetList.end()))
		{
			if (numSets > m_numSets)
			{
				Target * target = *it;
				{
					Progress::cleanup();
					char buf[80];
					sprintf(buf, "--- %s%d[%d,%d] -------------------\n", 
						Component::strDir(target->dir), target->dt, target->dx, target->dy);
//					Writer::print(buf);
					Progress::report();
				}
				searchTarget(target, pointSet, true);
			}
			++it;
		}
	}
	Progress::cleanup();
}

void Search::searchTarget(Target * target, PointSet * pointSet, bool solve)
{
	// pointSet actually points to an array of PointSet objects
	// note that it is long enough to accomodate target->dt

 	bool optimize = m_sorted || (m_maxCells != 0);

	// we need to prepare a stack long enough
	// so let's calculate the shortest component's dt

	long maxStackDepth;

	{
		long minDt = m_componentList[0]->getDt();
		long i = m_componentListLength - 1;
		do
		{
			long cdt = m_componentList[i]->getDt();
			if (minDt > cdt)
			{
				minDt = cdt;
			}
			--i;
		} while (i != 0);
		maxStackDepth = (target->dt + minDt - 1) / minDt;
	}

	// prepare the stack

	m_cc = new CompoundComponent(maxStackDepth, m_register->getMaxVariants(), maxStackDepth * m_register->getMaxVariants(), target->string);
	assert(m_cc != NULL);

	StackItem * stackArr = new StackItem[maxStackDepth + 1];
	assert(stackArr != NULL);

	StackItem * stack = stackArr;

	stack->nd = m_register->getAcTreeRoot();
	stack->compIdx = m_componentListLength - 1;
	stack->dt = target->dt;
	{
		long x;
		long y;
		Direction dir;
		Component::reverse(target->dx, target->dy, target->dir, &x, &y, &dir);
		m_cc->setPos(x, y, dir);
	}
	stack->x1 = m_cc->getCurX();
	stack->x2 = m_cc->getCurX();
	stack->y1 = m_cc->getCurY();
	stack->y2 = m_cc->getCurY();

	long level = 0;

	long maxComponentIdx = m_componentListLength;

AddComponent:
	++stack;
	++level;
	stack->compIdx = maxComponentIdx - 1;

SetComponent:
	// set the component
	{
		BasicComponent * comp = m_componentList[stack->compIdx];
		// check the Aho/Corasick tree
		stack->nd = stack[-1].nd->m_ptr[comp->getId()];
		if (NULL != stack->nd)
		{
			// check time
			stack->dt = stack[-1].dt - comp->getDt();
			if (0 <= stack->dt)
			{
				// check sizes
				// store current position
				long x = m_cc->getCurX();
				long y = m_cc->getCurY();
				Direction dir = m_cc->getCurDir();
				// add the component
				m_cc->addComponent(comp);
				// check position
				if (pointSet[stack->dt].checkPoint(m_cc->getCurX(), m_cc->getCurY(), m_cc->getCurDir()))
				{
					long x1, x2, y1, y2;
					// get component's minimum rectangle
					comp->getMinimumSize(x, y, dir, x1, x2, y1, y2);
					// calculate size
					// (note that m_maxcells might be turned to non-zero during the search)
					stack->x1 = (stack[-1].x1 < x1) ? stack[-1].x1 : x1;
					stack->x2 = (stack[-1].x2 > x2) ? stack[-1].x2 : x2;
					stack->y1 = (stack[-1].y1 < y1) ? stack[-1].y1 : y1;
					stack->y2 = (stack[-1].y2 > y2) ? stack[-1].y2 : y2;

					if ((m_maxCells == 0)
						|| (((stack->x2 - stack->x1) * (stack->y2 - stack->y1)) <= m_maxCells))
					{
						// the component seems to be OK
						// check if we filled the stack
						if (0 != stack->dt)
						{
							// the search could continue
							if (3 == level)
							{
								// under third level, call the other procedure
								++m_numSets;

								if (solve) 
								{
									if (m_numSets > m_ignoreSets)
									{
										Progress::setStep(m_numSets);

										continueSearch(pointSet, stack, maxComponentIdx, target->closed, optimize, target);

										// check for interrupts etc
										if (!m_run)
										{
											goto Finish;
										}
									}
								}
							}
							else
							{
								// continue inside here
								goto AddComponent;
							}
						}
						// yes, we are at target position
						// do the last check of the tree if necessary
						else if (!target->closed || checkAcTree(stack->nd)) 
						{
							long backout = m_cc->check(optimize);
							if (0 == backout)
							{
								if ((m_maxCells == 0) || (m_maxCells >= m_cc->getArea()))
								{
									bool ok = true;
									// all open results are OK;
									// closed results must be checked
									if ((!target->closed) || isResultSuitable(m_cc))
									{
										if (solve)
										{
											printResult(m_cc, target->sDir);
										}
									}
								}
							}
							else
							{
								// backout is not zero
								// one level will be backed out automatically
								// let's do the rest

								--backout;
								
								if (backout != 0)
								{
									do
									{
										m_cc->removeLastComponent();
										--stack;
										--level;
										--backout;
									} while (backout != 0);
								}
							}
						}
					}
				}
				m_cc->removeLastComponent();
			}
		}

	}

NextComponent:

	if (0 != stack->compIdx)
	{
		--stack->compIdx;
		if (target->closed && (1 == level))
		{
			--maxComponentIdx;
		}
		goto SetComponent;
	}
	// back out one level
	--stack;
	--level;
	if (stack != stackArr)
	{
		m_cc->removeLastComponent();
		goto NextComponent;
	}

Finish:
	delete stackArr;
	delete m_cc;
}

void Search::continueSearch(PointSet * pointSet, StackItem * stack, long maxComponentIdx, bool closed, bool optimize, Target * target)
{
	--maxComponentIdx;
	StackItem * stackBase = stack;

AddComponent:
	++stack;
	stack->compIdx = maxComponentIdx;

SetComponent:
	// set the component
	{
		BasicComponent * comp = m_componentList[stack->compIdx];
		// check the Aho/Corasick tree
		stack->nd = stack[-1].nd->m_ptr[comp->getId()];
		if (NULL != stack->nd)
		{
			// check time
			stack->dt = stack[-1].dt - comp->getDt();
			if (0 <= stack->dt)
			{
				// check sizes
				// store current position
				long x = m_cc->getCurX();
				long y = m_cc->getCurY();
				Direction dir = m_cc->getCurDir();
				// add the component
				m_cc->addComponent(comp);
				// check position
				if (pointSet[stack->dt].checkPoint(m_cc->getCurX(), m_cc->getCurY(), m_cc->getCurDir()))
				{
					long x1, x2, y1, y2;
					// get component's minimum rectangle
					comp->getMinimumSize(x, y, dir, x1, x2, y1, y2);
					// calculate size
					// (note that m_maxcells might be turned to non-zero during the search)
					stack->x1 = (stack[-1].x1 < x1) ? stack[-1].x1 : x1;
					stack->x2 = (stack[-1].x2 > x2) ? stack[-1].x2 : x2;
					stack->y1 = (stack[-1].y1 < y1) ? stack[-1].y1 : y1;
					stack->y2 = (stack[-1].y2 > y2) ? stack[-1].y2 : y2;

					if ((m_maxCells == 0)
						|| (((stack->x2 - stack->x1) * (stack->y2 - stack->y1)) <= m_maxCells))
					{
						// the component seems to be OK
						// check if we filled the stack
						if (0 != stack->dt)
						{
							goto AddComponent;
						}
						// yes, we are at target position
						// do the last check of the tree if necessary
						if (!closed || checkAcTree(stack->nd)) 
						{
							long backout = m_cc->check(optimize);
							if (0 == backout)
							{
								if ((m_maxCells == 0) || (m_maxCells >= m_cc->getArea()))
								{
									bool ok = true;
									if ((!closed) || isResultSuitable(m_cc))
									{
										printResult(m_cc, target->sDir);
									}
								}
							}
							else
							{
								// backout is not zero
								// one level will be backed out automatically
								// let's do the rest
								--backout;
								if (backout != 0)
								{
									do
									{
										m_cc->removeLastComponent();
										--stack;
										if (stack == stackBase)
										{
											goto Finish;
										}
										--backout;
									} while (backout != 0);
								}
							}

							// check for interrupts etc
							if (!m_run)
							{
								goto Finish;
							}
						}
					}
				}
				m_cc->removeLastComponent();
			}
		}
	}

NextComponent:

	if (0 != stack->compIdx)
	{
		--stack->compIdx;
		goto SetComponent;
	}
	// back out one level
	--stack;
	if (stack != stackBase)
	{
		m_cc->removeLastComponent();
		goto NextComponent;
	}

Finish:
	;
}

// memory-cheap search
void Search::runCheapWay()
{
	// calculate number of sets

	m_numSets = 0;

	Writer::print("Counting component sets...\n");

	{
		TargetList::iterator it = m_targetList.begin();
		while (it != m_targetList.end())
		{
			searchTarget(*it, false);
			++it;
		}
	}

	char buf[80];
	sprintf(buf, "Found %d component sets\n", m_numSets);
	Writer::print(buf);

	if (m_ignoreSets == 0)
	{
		Writer::print("Searching...\n");
	}
	else
	{
		sprintf(buf, "Searching (starting at set %d) ...\n", m_ignoreSets + 1);
		Writer::print(buf);
	}

	Progress::activate(m_numSets);
	long numSets = m_numSets;
	m_numSets = 0;

	// then run the search for all lengths
	{
		TargetList::iterator it = m_targetList.begin();
		while (m_run && (it != m_targetList.end()))
		{
			if (numSets > m_numSets)
			{
				Target * target = *it;
				{
					Progress::cleanup();
					char buf[80];
					sprintf(buf, "--- %s%d[%d,%d] -------------------\n", 
						Component::strDir(target->dir), target->dt, target->dx, target->dy);
//					Writer::print(buf);
					Progress::report();
				}
				searchTarget(target, true);
			}
			++it;
		}
	}

	// erase the report line
	Progress::cleanup();
}

void Search::searchTarget(Target * target, bool solve)
{
	// first task:
	// find all subsets of selected components
	// which create a given length

	long * activeSet = new long[m_componentListLength];
	assert(activeSet != NULL);
	bool * swaps = new bool[m_componentListLength];
	assert(swaps != NULL);
	bool * turns = new bool[m_componentListLength];
	assert(turns != NULL);

	long numSwaps = 0;
	long numTurns = 0;

	switch (target->dir)
	{
	case TLX:
	case TRX:
		numSwaps = 1;
		// fall down
	case TL:
	case TR:
		numTurns = 1;
		break;
	case TFX:
	case TBX:
		numSwaps = 1;
	}

	for (int i = 0; i < m_componentListLength; i++)
	{
		activeSet[i] = 0;
		switch (m_componentList[i]->getDirection())
		{
		case TF:
		case TB:
			swaps[i] = false;
			turns[i] = false;
			break;

		case TFX:
		case TBX:
			swaps[i] = true;
			turns[i] = false;
			break;

		case TL:
		case TR:
			swaps[i] = false;
			turns[i] = true;
			break;

		case TLX:
		case TRX:
			swaps[i] = true;
			turns[i] = true;
			break;
		}
	}

	long setPtr = 0;
	long length = target->dt;
	m_run = m_componentListLength > 0;

	while (m_run)
	{
		// getting a new component
		// first try the maximum number of them
		long dt = m_componentList[setPtr]->getDt();
		long num = length / dt;
		activeSet[setPtr] = num;
		// subtract the sum of components
		length -= num * dt;
		if (swaps[setPtr])
		{
			numSwaps += num;
		}
		if (turns[setPtr])
		{
			numTurns += num;
		}

		// filled?
		if (length == 0)
		{
			// permute if the selected components have chance to make it
			if (((numSwaps & 1) == 0)
				&& ((numTurns & 1) == 0))
			{
				++m_numSets;

				if (solve) 
				{
					if (m_numSets > m_ignoreSets)
					{
						Progress::setStep(m_numSets);

						permute(activeSet, target);
					}
				}
			}
			// continue to backtrack
		}
		else
		{
			++setPtr;
			// end of field?
			if (setPtr < m_componentListLength)
			{
				continue;
			}
			// we're at the last component
			// there's no reason to backtrack it by 1
			// just decrement it back to zero
			--setPtr;

		}
		// backtrack
		// on last component, decrease to zero
		if (setPtr == m_componentListLength - 1)
		{
			num = activeSet[setPtr];

			length += num * m_componentList[setPtr]->getDt();

			if (swaps[setPtr])
			{
				numSwaps -= num;
			}
			if (turns[setPtr])
			{
				numTurns -= num;
			}

			activeSet[setPtr] = 0;
		}
		// decrease setPtr until we get nonzero count
		while (activeSet[setPtr] == 0)
		{
			if (setPtr == 0)
			{
				// we're finished
				goto finished;
			}
			--setPtr;
		}

		// decrease the number
		--activeSet[setPtr];
		if (swaps[setPtr])
		{
			--numSwaps;
		}
		if (turns[setPtr])
		{
			--numTurns;
		}

		length += m_componentList[setPtr]->getDt();
		// and next time continue with the next component
		++setPtr;
	}

finished:
	delete [] activeSet;
	delete [] swaps;
	delete [] turns;
}

void Search::permute(long *activeSet, Target * target)
{
	// in the activeSet, there are numbers of components
	// which make the required length together
	// now we have to test if they can be combined 
	// to get to the specified point at the specified orientation

	long i;
	long sumVariants = 0;

	m_setSize = 0;

	for (i = 0; i < m_componentListLength; i++)
	{
		if (activeSet[i] != 0)
		{
			m_setSize++;
		}
	}

	// OK, it seems to be possible

	// we'reprobably going to spend some time here
	// so let's prepare some structures first

	m_set = new CompSet[m_setSize + 1];
	assert(m_set != NULL);
	m_set[m_setSize].comp = NULL; // sentinel

	CompSet * setPtr = m_set;

	CompSet * best = m_set;
	m_setCount = 0;
	m_setSumLen = 0;

	for (i = 0; i < m_componentListLength; i++)
	{
		long n = activeSet[i];
		if (n != 0)
		{

			BasicComponent * comp = m_componentList[i];

			setPtr->num = n;
			setPtr->comp = comp;
			setPtr->length = sqrt(comp->getDx() * comp->getDx()
				+ comp->getDy() * comp->getDy() +0.0) + 1;

			m_setSumLen += setPtr->length * n;

			// calculate the least counted
			if (best->num > setPtr->num)
			{
				best = setPtr;
			}
			// increase the number of components
			++setPtr;
			// also calculate total number of components
			m_setCount += n;
			// and total number of variants
			sumVariants += n * comp->getNumVariants();
		}
	}

	// prepare result container of appropriate size

	m_cc = new CompoundComponent(m_setCount, m_register->getMaxVariants(), sumVariants, target->string);
	assert(m_cc != NULL);

	// create all permutations
	// start with the found component
	// (if it's on a closed track, of course)
	scanPermutations(best, target);

	delete [] m_set;
	delete m_cc;
}

void Search::scanPermutations(CompSet * best, Target * target)
{
	// this procedure is extremely ugly, particularly due to its use of goto statements
	// however, the search spends a great amount of time here

 	bool optimize = m_sorted || (m_maxCells != 0);

	typedef struct {
		// A/C Tree position for this stack depth
		AcTreeNode * nd;
		// the component used at this level
		BasicComponent * comp;
		// Current position and direction
		long dx;
		long dy;
		Direction dir;
		// Current search pointer
		CompSet * setPtr;
	} StackItem;

	StackItem * stackArr = new StackItem[m_setCount + 1];
	assert(stackArr != NULL);
	StackItem * stack = stackArr;
	StackItem * stkBegin; // pointer at one stack position before the first valid component

	stack->nd = m_register->getAcTreeRoot();
	Component::reverse(target->dx, target->dy, target->dir, &stack->dx, &stack->dy, &stack->dir);

	stack->comp = NULL; // the bottom level of stack doesn't contain a component

	stack->setPtr = NULL;

	// if we search a closed loop, let's use the 'best' component first
	if (target->closed)
	{
		// implant the best component instead of the sentinel

		stack->comp = best->comp;
		--best->num;
		--m_setCount;
		m_setSumLen -= best->length;

		stack->nd = stack->nd->m_ptr[stack->comp->getId()];

		stack->dx += stack->comp->getDx(stack->dir);
		stack->dy += stack->comp->getDy(stack->dir);
		stack->dir = stack->comp->getDirection(stack->dir);

		// set the set pointer so that we don't try another component at this depth
		stack->setPtr = best;

		// the component list begins here
		stkBegin = stack - 1;
	}
	else
	{
		// the component list begins at the next position
		stkBegin = stack;
	}

	long uncheckedDepth = 0;

AddComponent:

	if (!m_run)
	{
		goto Finish;
	}

	// stack position is filled

	// move to a new position
	++stack;
	// and initialise the setPtr
	stack->setPtr = m_set;

	++uncheckedDepth;

SetComponent:

	// stack position is empty

	// find the first available component
	if (stack->setPtr->num == 0)
	{
		do
		{
			stack->setPtr++;
			if (stack->setPtr->comp == NULL)
			{
				goto BackOut;
			}
		} while (stack->setPtr->num == 0);
	}

	// set the component
	stack->comp = stack->setPtr->comp;

	// check the tree
	stack->nd = stack[-1].nd->m_ptr[stack->comp->getId()];

	if (stack->nd == NULL)
	{
		goto NextComponent;
	}

	// component is valid
	{
		stack->setPtr->num--;
		--m_setCount;
		m_setSumLen -= stack->setPtr->length;

		stack->dir = stack->comp->getDirection(stack[-1].dir);
		stack->dx = stack->comp->getDx(stack[-1].dir) + stack[-1].dx;
		stack->dy = stack->comp->getDy(stack[-1].dir) + stack[-1].dy;
	}

	// now the stack position is filled

	if (m_setCount != 0)
	{
		// check the distance to the target
		if (stack->dx * stack->dx + stack->dy * stack->dy <= m_setSumLen * m_setSumLen)
		{
			// good for now
			// we can make it, try more components
			goto AddComponent;
		}
		else
		{
			if (uncheckedDepth > 8)
			{
				m_cc->reset();
				StackItem * stk = stkBegin;
				while (stk != stack)
				{
					++stk;
					m_cc->addComponent(stk->comp);
				}

				uncheckedDepth = 0;
				long backout = m_cc->check(false);
				if (backout > 1)
				{
					--backout;
					do
					{
						// stack position is full
						++m_setCount;
						m_setSumLen += stack->setPtr->length;
						stack->setPtr->num++;
						// stack position is empty

						--stack;

						--backout;
					} while (backout != 0);
				}
			}
		}
	}
	else
	{
		// all components are in place
		// check if we're at taget
		if ((stack->dx == 0)
			&& (stack->dy == 0)
			&& (stack->dir == TF)
			)
		{
			// it looks good
			// let's put all the components into the compound component
			// for further checks
			m_cc->reset();
			StackItem * stk = stkBegin;
			while (stk != stack)
			{
				++stk;
				m_cc->addComponent(stk->comp);
			}

			// check the tree and minimum size
			if (!target->closed || checkAcTree(stack->nd)) 
			{
				long backout;

				if (m_maxCells != 0)
				{
					backout = m_cc->checkMinArea(m_maxCells);
					if (backout == 0)
					{
						uncheckedDepth = 0;
						backout = m_cc->check(optimize);
					}
					else
					{
						if (backout == m_cc->getNumComponents())
						{
							// even the first component doesn't fit
							// fall out if the first component is fixed
							if (stkBegin != stackArr)
							{
								// closed loop; first component is fixed
								// and it is too big
								goto Finish;
							}
						}
					}
				}
				else
				{
					uncheckedDepth = 0;
					backout = m_cc->check(optimize);
				}

				if (backout == 0)
				{
					if ((m_maxCells == 0) || (m_maxCells >= m_cc->getArea()))
					{
						printResult(m_cc, target->sDir);
					}
				}
				else
				{
					if (--backout > 0)
					{
						do
						{
							++m_setCount;
							m_setSumLen += stack->setPtr->length;
							stack->setPtr->num++;

							// stack position is empty

							--stack;
							if (uncheckedDepth != 0)
							{
								--uncheckedDepth;
							}

							// stack position is full
						} while (--backout != 0);
					}
				}
			}
		}
	}

ResetAndNext:

	// stack position is full

	// remove the last component

	++m_setCount;
	m_setSumLen += stack->setPtr->length;
	stack->setPtr->num++;

	// stack position is empty

NextComponent:

	// stack position is empty

	stack->setPtr++;
	if (stack->setPtr->comp != NULL)
	{
		goto SetComponent;
	}

	// no next component available

BackOut:

	// stack position is empty

	--stack;

	if (uncheckedDepth != 0)
	{
		--uncheckedDepth;
	}

	if (stack != stackArr)
	{
		// stack position is full

		goto ResetAndNext;
	}

Finish:

	delete [] stackArr;
}

void Search::printResult(CompoundComponent * comp, Direction dir)
{
	Progress::cleanup();

	printComponent(comp, m_numSets);

	if (m_sorted)
	{
		m_resultList.push(comp->clone(dir));
		if (m_optimizeMaxCells && (m_resultList.size() > 2 * m_maxResults + 1))
		{
			std::deque<CompoundComponent *> qq;

			long s = 0;

			while (!m_resultList.empty())
			{
				CompoundComponent * cc = m_resultList.top();
				m_resultList.pop();

				if (qq.size() < m_maxResults)
				{
					s = cc->getArea(); // remember the size
					qq.push_back(cc);
				}
				else
				{
					// discard
					delete cc;
					cc = NULL;
				}
			}

			while (!qq.empty())
			{
				m_resultList.push(qq.front());
				qq.pop_front();
			}

			m_maxCells = s;

			char buf[80];
			sprintf(buf, "+++ Maximum cells limit set to %d\n", m_maxCells);
			Writer::print(buf);
		}
	}
	else
	{
		CompoundComponent * cc = comp->clone(dir);
		cc->printRle(!m_closed);
		delete cc;

		if (m_maxResults != 0)
		{
			--m_maxResults;
			if (m_maxResults == 0)
			{
				m_run = false;
				Writer::print("Result number limit reached\n");
			}
		}
	}

	Progress::report();

	Writer::flush();
}

void Search::printSorted()
{
	if (!m_resultList.empty())
	{
		Writer::print("... Sorting ...\n");

		bool storeRle = true;

		while(!m_resultList.empty())
		{
			CompoundComponent * cc = m_resultList.top();
			m_resultList.pop();
			printComponent(cc, 0);
			if (storeRle)
			{
				cc->printRle(!m_closed);
				if (m_maxResults != 0)
				{
					--m_maxResults;
					if (m_maxResults == 0)
					{
						storeRle = false;
						Writer::print("\nResult number limit reached\n\n");
					}
				}
			}
			delete cc;
		}
		Writer::flush();
	}
}

void Search::printComponent(CompoundComponent *comp, long setNo)
{
	Writer::print("*** ");
	comp->print(setNo);
	Writer::print("\n");
}

void Search::interrupt()
{
	m_run = false;
	m_interrupted = true;
}

bool Search::checkAcTree(AcTreeNode *nd)
{
	// the component in m_cc is a closed loop
	// and nd points to the point in the A/C tree
	// where we got from the start

	// calculate number of components to examine
	// either maximum tree depth
	// or number of components in loop
	// smaller of the two

	long m = m_cc->getNumComponents();
	if (m > AcTree::maxAcTreeDepth)
	{
		m = AcTree::maxAcTreeDepth;
	}

	long i = 0;
	do
	{
		AcTreeNode * ndd = nd->m_ptr[m_cc->getComponentId(i)];
		if (ndd == NULL)
		{
			// either we have a problem
			// or we closed the loop
			// depends on how deep the node is

			return nd->m_depth == m_cc->getNumComponents();
		}
		nd = ndd;
	} while (++i < m);

	return true;
}

void Search::addTarget(long dx1, long dy1, long dt1, Direction dir1, long dx2, long dy2, long dt2, Direction dir2)
{
	Direction dir;
	long dx, dy, dt;

	dt = dt2 - dt1;
	dir = Component::decombine(dir1, dir2);
	Component::globalToLocal(dx1, dy1, dir1, dx2, dy2, &dx, &dy);

	addTarget(dx, dy, dt, dir, dir1);
}

void Search::addTarget(long dx1, long dy1, long dt1, Direction dir1, long dx2, long dy2, long dt2, Direction dir2, char *str)
{
	Direction dir;
	long dx, dy, dt;

	dt = dt2 - dt1;
	dir = Component::decombine(dir1, dir2);
	Component::globalToLocal(dx1, dy1, dir1, dx2, dy2, &dx, &dy);

	addTarget(dx, dy, dt, dir, dir1, str);
}

void Search::addTarget(long dx, long dy, long dt, Direction dir)
{
	addTarget(dx, dy, dt, dir, TF);
}

void Search::addTarget(long dx, long dy, long dt, Direction dir, Direction startDir)
{
	if (dt > 0)
	{
		Target * tg = new Target(dx, dy, dt, dir, startDir);
		assert(tg != NULL);
		if (!tg->closed)
		{
			// if any of targets isn't closed
			// then we are generally searching for open tracks
			m_closed = false;
		}
		if (!m_targetList.insert(tg).second)
		{
			delete tg;
		}
	}
}

void Search::addTarget(long dx, long dy, long dt, Direction dir, Direction startDir, char *str)
{
	if (dt > 0)
	{
		Target * tg = new Target(dx, dy, dt, dir, startDir, str);
		assert(tg != NULL);
		if (!tg->closed)
		{
			// if any of targets isn't closed
			// then we are generally searching for open tracks
			m_closed = false;
		}
		if (!m_targetList.insert(tg).second)
		{
			delete tg;
		}
	}
}

bool Search::isResultSuitable(CompoundComponent * cc)
{
	// check all sequences starting with the same component as the first one
	// if a different sequence represents 'smaller number', the result is not suitable
	long n = cc->getNumComponents();
	for (long i = 1; i < n; ++i)
	{
		if (cc->getComponentId(i) == cc->getComponentId(0))
		{
			for (long j = 1; j < n; ++j)
			{
				if (cc->getComponentId((i + j) % n) < cc->getComponentId(j))
				{					
					return false;
				}
				else if (cc->getComponentId((i + j) % n) > cc->getComponentId(j))
				{
					break;
				}
			}
		}
	}
	return true;
}

void Search::trimTargets(long maxDt)
{
	long n = 0;
	long d = 0;
	TargetList::iterator it = m_targetList.begin();
	while ((it != m_targetList.end()) && ((*it)->dt <= maxDt))
	{
		++it;
		++n;
	}

	while (it != m_targetList.end()) 
	{
		it = m_targetList.erase(it);
		++d;
	}

	if (d == 0)
	{
		char buf[80];
		sprintf(buf, "Accepted %d target%s\n", n, (n == 1) ? "" : "s");
		Writer::print(buf);
	}
	else
	{
		char buf[80];
		sprintf(buf, "Accepted %d target%s; %d target%s over t=%d ignored\n", n, (n == 1) ? "" : "s", d, (d == 1) ? "" : "s", maxDt);
		Writer::print(buf);
	}
}
